proc ucvtbp {b} {
	return "[lindex $b 0]:[lindex $b 1].[lindex $b 2]"
}

proc cvtbp {line} {
	set char 0
	set file ""
	if {[string match *:* $line]} {
		set file [shpat_% $line :.*]
		set line [shpat_# $line $file:]
	}
	if {[string match *.* $line]} {
		set char [shpat_## $line {.*\.}]
		set line [shpat_% $line "\\.$char"]
	}
	if {$line == ""} {set line 0}
	if {$char == ""} {set char 0}
	return [list $file $line $char]
}

proc listbp {point} {
	set bp [cvtbp $point]
	set file [lindex $bp 0]
	set line [lindex $bp 1]
	set char [lindex $bp 2]
	set l [deet_breakpoint -l $file $line $char]
	if {[llength $l] == 0} {
		error "No breakpoint on line $line char $char"
	} elseif {[llength $l] > 1} {
		error "Select one of $l"
	}
	return [lindex $l 0]
}

proc getloc {} {
	global DeetSeqNum DeetFrame DeetLocation DeetGraphical

        incr DeetSeqNum
	if {[catch {set f [deet_frame]}]} {
		debugwin_bp show "" 0 0
		return ""
	}
        set file [lindex $f 2]
        set line [lindex $f 3]
        set char [lindex $f 4]
        set DeetFrame [lindex $f 0]
        set DeetLocation [list $file $line $char]
	if {($file == "") || ($line == 0)} {return $file:$line.$char}
	if {$DeetGraphical} {
		debugwin_bp show $file $line $char
		vlist_update
		xwhere_stack
	} else {
		puts "Stopped at file $file, line $line."
		puts -nonewline "$line: "
		exec head -$line $file | tail -1
	}
	return $file:$line.$char
}

proc deetgo {w {cmd deet_continue}} {
	global breakpoint

	if {$w != ""} {message_set $w.message "Running..."}
	eval $cmd

	set bp [getloc]
	if {$bp == ""} {return 1}
	if {! [sh_vnull breakpoint($bp)]} {
		catch {if {! [eval $breakpoint($bp)]} {return 0}}
	}

	return 1
}

proc deetgoto {w bp} {
	global breakpoint
	foreach b [array names breakpoint] {
		set savedBreakpoint($b) breakpoint$(b)
		delete $b
	}
	b $bp
	deetgo $w "deet_continue -c"
	delete $bp
	foreach b [array names savedBreakpoint] {
		b $b
	}
}

proc cont {{w ""}} {
	while {[deetgo $w "deet_continue -c"] == 0} {}
}

proc step {{w ""}} {
	deetgo $w "deet_continue -s"
}

proc next {{w ""}} {
	deetgo $w "deet_continue -n"
}

proc stop {{w ""}} {
	deet_continue -i
	getloc
}

proc b {point {func ""}} {
	global DeetGraphical breakpoint
	if {[catch {listbp $point} bp]} {
		return $bp
	} else {
		eval deet_breakpoint -s $bp
		set breakpoint([ucvtbp $bp]) $func
		if {$DeetGraphical} {	
			eval debugwin_bp on $bp
			eval bpwin_bp on $bp
		}
	}
}

proc delete {point} {
	global DeetGraphical breakpoint
	if {[catch {listbp $point} bp]} {
		return $bp
	} else {
		set point [ucvtbp $bp]
		if {! [info exists breakpoint($point)]} {
               		error "Breakpoint not set"
		}
		eval deet_breakpoint -d $bp
		unset breakpoint($point)
		if {$DeetGraphical} {
			eval debugwin_bp off $bp
			eval bpwin_bp off $bp
		}
	}
}

proc disable {point} {
	global breakpoint disabled
	if {[catch {listbp $point} bp]} {
		return $bp
	} else {
		set point [ucvtbp $bp]
		if {[info exists breakpoint($point)]} {
			set disabled($point) $breakpoint($point)
			delete $point
			eval bpwin_bp on $bp
			eval debugwin_bp dis $bp
		}
	}
}

proc enable {point} {
	global breakpoint disabled
	if {[catch {listbp $point} bp]} {
		return $bp
	} else {
		set point [ucvtbp $bp]
		if {[info exists disabled($point)]} {
			unset disabled($point)
			b $point
		}
	}
}

proc bpwin_bp {mode file line char} {
	global disabled
	set w .breakpoints

	if {! [winfo exists $w]} return
	set bp "$file:$line.$char"
	if {$mode == "on"} {
		regsub / $file | id
		hlist_insert $w.bp $id "File $file"
		hlist_insert $w.bp $id/$line "Line $line char $char" $bp
		if {[info exists disabled($bp)]} {
			[hlist_subwidget $w.bp text] tag configure \
				$id/$line -foreground red
		}
	} else {
		if {[info exists disabled($bp)]} return
		regsub / $file | id
		hlist_delete $w.bp $id/$line
	}
}

proc xbreakpoints {} {
	global breakpoint disabled

	set w .breakpoints
	if {[winfo exists $w]} return

	toplevel $w
	window_title $w "Set Breakpoints"
	menu_create $w.menu File:l
	menu_addcommand $w.menu 1 Load "xbreakpoints_action $w load"
	menu_addcommand $w.menu 1 Save "xbreakpoints_action $w save"
	menu_addcommand $w.menu 1 Close "xbreakpoints_action $w close"
	hlist_create $w.bp
	buttons_create $w.but Close:r Delete:r Enable:r Disable:r Goto:r
	buttons_command $w.but 1 "xbreakpoints_action $w close"
	buttons_command $w.but 2 "xbreakpoints_action $w delete"
	buttons_command $w.but 3 "xbreakpoints_action $w enable"
	buttons_command $w.but 4 "xbreakpoints_action $w disable"
	buttons_command $w.but 5 "xbreakpoints_action $w goto"
	mpack $w.menu $w.but:bx $w.bp:ez 

	set t [hlist_subwidget $w.bp text]
	foreach bp [concat [array names breakpoint] [array names disabled]] {
		eval bpwin_bp on [cvtbp $bp]
	}
}

proc xbreakpoints_action {w action} {
	global breakpoint disabled

	set h $w.bp
	set sel [hlist_selection $h]
	set name [lindex $sel 0]
	set bp [lindex $sel 1]
	set t [hlist_subwidget $h text]

	switch $action {
		disable {
			if {$bp != "" && [info exists breakpoint($bp)]} {
				set disabled($bp) $breakpoint($bp)
				delete $bp
				$t tag configure $name -foreground red
			}
		}
		enable {
			if {$bp != "" && [info exists disabled($bp)]} {
				unset disabled($bp)
				$t tag configure $name -foreground black
				b $bp
			}
		}
		close {
			destroy $w
		}
		delete {
			if {$bp != ""} {
				hlist_delete $h $name
				delete $bp
			}
		}
		edit {
			set c [window_create]
			text_create $c.t
			pack $c.t
			text_set $c.t "Not implemented"
			# set breakpoint($data) [text_get $t]
		}
		load {
		        set file [getfile_open "File Selection"]
			if {$file != ""} {source $file}
		}
		save {
		        set file [getfile_save "File Selection"]
			if {$file != "" && ! [file exists $file]} {
				puts "breakpoints not saved: not implemented"
			}
		}
		goto {
			if {$bp != ""} {
				eval debugwin_bp show [cvtbp $bp]
			}
		}
	}
}

proc xbreakpoints_setcond {bp} {
}

proc breakpoints {} {
	global breakpoint
	set result ""
	foreach i [array names breakpoint] {
		set bp [cvtbp $i]
		set f [lindex $f 0]
		set l [lindex $f 1]
		set c [lindex $f 2]
		append result "File $f, line $l char $c\n"
	}
	return $result
}

proc opentarget {name {dargs ""}} {
	global DeetSource
	deet_open $name $dargs
	foreach w [window_list Debugger] {
		debugwin_destroy $w
	}
	catch {unset DeetSource}
}

proc target {} {
	set file [getfile_open "Location of target file:"]
	if {$file != ""} {
		cd [file dirname $file]
		opentarget $file ""
		set w [debugwin_create]
	        catch {debugwin_setfile $w [findfunc main]}
	}
}
